package Texitheque::CGI;

use strict;
use warnings;

use Texitheque::CGI::StockMessages;

use CGI qw/-no_xhtml -nosticky/;
use CGI::Carp;
use File::Spec::Functions qw/:ALL/;
use File::Basename;

sub new {
  my $this = shift;
  my $class = ref $this || $this;
  my $self = {};

  # Read valid options
  if (@_) {
    my $valid_opts = $class->_valid_module_options;
    return unless ref $valid_opts eq 'HASH';

    while (@_) {
      my $key = shift;

      if (ref $key eq 'HASH') {
	while (my ($hkey, $hval) = each %$key) {
	  unless ($valid_opts->{$hkey}) {
	    carp "$class: Unknown option: '$hkey'\n";
	    next;
	  }
	  $self->{$hkey} = $hval;
	}
      } else {
	unless ($valid_opts->{$key}) {
	  carp "$class: Unknown option: '$key'\n";
	  shift; next;
	}
	$self->{$key} = shift;
      }
    }
  }

  # Check if required options are provided
  my $req_opts = $class->_required_module_options;
  return unless ref $req_opts eq 'ARRAY';

  my $unsatisfied;

  foreach (@$req_opts) {
    unless (defined $self->{$_}) {
      carp "$class option '$_' must be set\n";
      $unsatisfied++;
    }
  }

  return if $unsatisfied;

  bless $self, $class;

  return unless $self->initialize;
  
  return $self;
}

sub initialize {
  my $self = shift;

  $self->{q} = CGI->new;

  print $self->{q}->header;

  # Get webmaster address ready for HTML
  $self->{webmaster} = $self->{q}->a({-href => "mailto:$self->{-webmaster}"}, 'webmaster');

  # Unusable template file
  unless (-r $self->{-template}) {
    carp "Template file '$self->{-template}' unreadable\n";
    print $self->message($Texitheque::CGI::StockMessages::Template_Error, {'webmaster' => $self->{webmaster}});
    return;
  }

  # Get the template parsed and ready
  if (open TEMPLATE, $self->{-template}) {
    my $template = '';

    while (<TEMPLATE>) {
      chomp; $template .= $_;
    }

    $self->{_template_start} = $self->{_template_end} = '';

    if ($template =~ m/^(.*)\@CGITEMPLATE\@(.*)$/) {
      $self->{_template_start} = $1;
      $self->{_template_end}   = $2;
    }

    close TEMPLATE;
  } else {
    carp "Cannot open('$self->{-template}'): $!\n";
    print $self->message($Texitheque::CGI::StockMessages::Template_Error, {'webmaster' => $self->{webmaster}});
    return;
  }

  # $self->{-mime_images} for filesystem checks
  # $self->{_mime_images} for mime image links in HTML
  if (defined $self->{-mime_images} and -d $self->{-mime_images}) {
    if (defined $self->{-web_base_rel}) {
      $self->{_mime_images} = catdir
	$self->{-web_base_rel},
	abs2rel($self->{-mime_images}, $self->{-web_base});
    } else {
      $self->{_mime_images} = abs2rel($self->{-mime_images}, $self->{-web_base});
    }
  }

  # $self->{-thumbs} for filesystem works
  # $self->{_thumbs} for thumbnail links in HTML
  if (defined $self->{-thumbs}) {
    if (defined $self->{-web_base_rel}) {
      $self->{_thumbs} = catdir
	$self->{-web_base_rel},
	abs2rel($self->{-thumbs}, $self->{-web_base});
    } else {
      $self->{_thumbs} = abs2rel($self->{-thumbs}, $self->{-web_base});
    }
  }

  $self->get_cgi_vars;
  $self->set_display_mode;

  1;
}

sub get_cgi_vars {
  my $self = shift;
  my $valid_options = $self->_cgi_vars;

  return unless ref $valid_options eq 'ARRAY';

  foreach my $opt (@$valid_options) {
    $self->{$opt} = $self->{q}->param($opt);
  }
}

sub run {
  my $self = shift;

  $self->_start;

  $self->print_stuff if $self->gather_stuff;

  $self->_finalize;
}

sub start {
  my $self = shift;

  unless ($self->{_template_started}) {
    print $self->{_template_start};
    $self->{_template_started}++;
    return 1;
  }

  return;
}

sub _start {
  my $self = shift;

  $self->{_template_auto}++ if $self->start;
}

sub finalize {
  my $self = shift;

  if ($self->{_template_started} and not $self->{_template_ended}) {
    print $self->{_template_end};
    $self->{_template_ended}++;
    return 1;
  }

  return;
}

sub _finalize {
  my $self = shift;

  $self->finalize if $self->{_template_auto};
}

sub clean_path {
  my ($self, $path) = @_;

  return unless defined $path;

  return canonpath catdir no_upwards splitdir $path;
}

sub Dump {
  my $self = shift;

  use Data::Dumper;
  use HTML::Entities;

  print
    $self->{q}->pre(encode_entities(Dumper($self))),
    $self->{q}->hr,
    $self->{q}->Dump,
}

sub set_display_mode {
  my $self = shift;

  my $valid_modes = $self->_valid_display_modes;

  return unless ref $valid_modes eq 'HASH';

  if (defined $self->{m} and defined $valid_modes->{$self->{m}}) {
    return ($self->{_mode} = $self->{m});
  }

  if (defined $self->{-default_mode}) {
    if (defined $valid_modes->{$self->{-default_mode}}) {
      return ($self->{_mode} = $self->{-default_mode});
    } else {
      carp 'Unknown default mode for ' . ref($self) . '. Must be one of: ' . join(' ', keys %$valid_modes) . "\n";
    }
  }

  my $api_default_mode = $self->_default_mode;

  if (defined $api_default_mode and defined $valid_modes->{$api_default_mode}) {
    return ($self->{_mode} = $api_default_mode);
  } else {
    croak 'Danger: Inconsistent display mode subs for ' . ref($self) . "\n";
  }
}

sub mode_switcher {
  my $self = shift;

  my $modes = $self->_valid_display_modes;

  return '' unless ref $modes eq 'HASH';

  my $q = $self->{q};

  my $switcher = $q->strong('Switch browsing mode:');

  my $url = $q->url(-relative => 1) . '?';

  my $opts = $self->_cgi_vars;
  if (defined $opts and ref $opts eq 'ARRAY' and scalar @$opts) {
    my $vars = join '&', map { $_ . '=' . $self->{$_} } grep { $_ ne 'm' and $self->{$_} } @$opts;
    $url .= $vars . '&' if $vars;
  }

  while (my ($mode, $tip) = each (%$modes)) {
    # Spacer
    $switcher .= '&nbsp;' x 4;

    if ($self->{_mode} eq $mode) {
      $switcher .= $q->span({-title => $tip}, $mode);
    } else {
      $switcher .= $q->a({-href => $url . "m=$mode", -title => $tip}, $mode);
    }
  }

  return $switcher;
}

sub deuglify_size {
  my ($self, $size) = @_;

  return unless defined $size;

  # 1 kilobyte = 1024 bytes
  my $kb = 1024;
  # 1 megabyte = 1024 kilobytes
  my $mb = 1024 * $kb;

  # Quick file size deuglification, ie. 18600 => 18 KBytes, 766 => 766 Bytes
  if ($size < $kb) {
    $size .= ' Bytes';
  } elsif ($size <$mb) {
    $size = $size/$kb;
    $size = sprintf('%.1f', $size);
    $size .= ' KBytes';
  } else {
    $size = $size/$mb;
    $size = sprintf('%.2f', $size);
    $size .= ' MBytes';
  }

  return $size;
}

sub get_file_size {
  my ($self, $file) = @_;

  return unless $file;

  # File size in bytes
  my $size = (stat($file))[7];

  return $self->deuglify_size($size);
}

sub get_dir_size {
  my ($self, $dir) = @_;

  return unless defined $dir;

  if (opendir DIR, $dir) {
    my $size = scalar grep { !/^\./ and !/\.DSC$/ } readdir DIR;

    close DIR;

    return $size . ' ' . ($size == 1 ? 'Entry' : 'Entries');
  }

  return;
}

sub readfile {
  my ($self, $file) = @_;

  return '' unless defined $file;
  return '' unless -r $file;

  open(FILE, "<$file");
  my @content = <FILE>;
  close(FILE);

  my $content = join('', @content);

  $content =~ s/\n*$/\n/;

  chomp $content;

  $content =~ s/\n/<br>\n/;

  return $content;
}

sub get_file_desc {
  my ($self, $file) = @_;

  return unless defined $file;
  return unless -r $file;

  # Information about file is maybe in $desc_file
  my $desc_file = $file . '.DSC';

  # If we can get information from $file.DSC, let's use it
  my $desc = $self->readfile($desc_file);
  return $desc if $desc;

  # Decide what to do according to the file extension
  SWITCH: {
    if ($file =~ /\.gslwave$/i) { return $self->get_gslwave_info($file); last SWITCH; }
    if ($file =~ /\.bse$/i)     { return $self->get_bse_info($file);     last SWITCH; }
  }

  # Shite, no info :\
  return '';
}

sub get_gslwave_info {
  my ($self, $file) = @_;

  return unless defined $file;
  return unless -r $file;

  if (open FILE, "<$file") {
    while (<FILE>) {
      chomp;
      return $2 if /name\s*=\s*(["'])([^\1]*)\1/;
    }
    close(FILE);
  } else {
    carp "Cannot open $file for reading: $!";
    return '';
  }

  # Damn, can't find anything :\
  return '';
}

sub get_bse_info {
  my ($self, $file) = @_;

  return unless defined $file;
  return unless -r $file;

  # FIXME
  # return '[ Reading file info from .bse files not yet implemented. ]';
  return ''
}

sub get_dir_desc {
  my ($self, $dir) = @_;

  return unless defined $dir;

  my $desc_file = catfile($dir, '.dir.DSC');
  
  return $self->readfile($desc_file);
}

sub get_dir_icon {
  my $self = shift;

  return '' unless $self->{_mime_images};

  my $image = 'dir.png';

  if (-f (catfile $self->{-mime_images}, $image)) {
    return catfile $self->{_mime_images}, $image;
  }

  return '';
}

sub get_file_icon {
  my ($self, $file) = @_;

  return '' unless $self->{_mime_images};
  return '' unless defined $file;

  $file =~ m/\.([^.]+)$/;
  my $ext = lc $1;

  if ($ext and -f (catfile $self->{-mime_images}, $ext . '.png')) {
    return catfile $self->{_mime_images}, $ext . '.png';
  }

  # Fallback to default, if it's available
  my $default = 'file.png';

  if (-f (catfile $self->{-mime_images}, $default)) {
    return catfile $self->{_mime_images}, $default;
  }

  return '';
}

sub get_file_link {
  my ($self, $file) = @_;

  return unless defined $file;

  if (defined $self->{-web_base_rel}) {
    return catdir $self->{-web_base_rel}, abs2rel($file, $self->{-web_base});
  } else {
    return abs2rel($file, $self->{-web_base});
  }
}

sub get_dir_link {
  my ($self, $dir) = @_;

  my $RE = qr#$self->{_base}#;

  $dir =~ s/^$RE//;

  # Remove slashes from the beginning
  $dir =~ s/^\/*//g;

  my @options;
  push @options, "d=$dir" if $dir;

  # Rest of options
  push @options, 'm=' . $self->{m} if $self->{m};

  my $url = $self->{q}->url(-relative => 1);
  $url .= '?' . join('&', @options) if @options;

  return $url;
}

sub get_date {
  my ($self, $file) = @_;

  return unless defined $file;
  return unless -r $file;

  my $mtime = localtime((stat $file)[9]);

  chomp (my $date = qx(date '+\%A, \%e \%b \%Y, \%H:\%M' --date='$mtime'));

  return $date;
}

sub get_table_title {
  my $self = shift;

  $self->{table_title} = $self->get_dir_desc($self->{_base});
}

sub message {
  my ($self, $t, $v)  = @_;

  return unless defined $t;
  return $t unless ref $v eq 'HASH';

  while (my ($key, $val) = each %$v) {
    $t =~ s/\@$key\@/$val/g;
  }

  return $t;
}

1;
