package Texitheque::CGI::FileBrowser;

use strict;
use warnings;
use base 'Texitheque::CGI';

use CGI::Carp;
use File::Spec::Functions qw/:ALL/;
use File::Basename;

sub gather_stuff {
  my $self = shift;

  $self->{_base} = canonpath $self->{-root_dir};

  $self->{_scan} = $self->clean_path($self->{d}) || '';

  $self->{_scan_fs} = canonpath catdir($self->{_base}, $self->{_scan});

  # Fail-safety.  Fallback to root scandir.
  $self->{_scan_fs} = $self->{_base} unless -d $self->{_scan_fs};

  # Initialize arrays
  $self->{contents}   = [];
  $self->{upper_dirs} = [];

  $self->get_table_title;
  $self->get_upper_directories;
  $self->scan_directory;
}

sub scan_directory {
  my $self = shift;

  unless (-d $self->{_scan_fs}) {
    carp "Cannot read '$self->{_scan_fs}' directory\n";
    print $self->message($Texitheque::CGI::StockMessages::Directory_Read_Error, {'error' => $!});
    return;
  }

  unless (opendir DIR, $self->{_scan_fs}) {
    carp "Cannot opendir('$self->{_scan_fs}'): $!\n";
    print $self->message($Texitheque::CGI::StockMessages::Directory_Read_Error, {'error' => $!});
    return;
  }

  foreach (grep { !/^\./ and !/\.DSC$/ } readdir DIR) {
    push @{$self->{contents}}, $self->get_entry_info(catfile $self->{_scan_fs}, $_);
  }
  closedir DIR;

  @{$self->{contents}} = sort _sort_dirs_files @{$self->{contents}};

  1;
}

sub get_upper_directories {
  my $self = shift;

  return unless $self->{_scan};

  if ($self->{_mode} eq 'tree') {
    $self->get_upper_directories_tree;
  } else {
    $self->get_upper_directories_standard;
  }
}

sub get_upper_directories_tree {
  my $self = shift;

  my $curdir  = $self->{_scan};
  my $prevdir = '';

  while ($curdir ne $prevdir and $curdir ne '.') {
    unshift @{$self->{upper_dirs}}, $self->get_entry_info(catfile $self->{_base}, $curdir);

    $prevdir = $curdir;
    $curdir  = dirname $curdir;
  }

  # Add a root item, and adjust it a little bit
  unshift @{$self->{upper_dirs}}, $self->get_entry_info($self->{_base});
  $self->{upper_dirs}->[0]->{name} = '/ (Root)';

  # No link for current directory
  delete $self->{upper_dirs}->[$#{$self->{upper_dirs}}]->{link};

  # Hacky but fast and it works :)
  # Last item is current directory, so make its icon 'dir-open.png'
  $self->{upper_dirs}->[$#{$self->{upper_dirs}}]->{icon} =~ s/dir\.png/dir-open\.png/;

  1;
}

sub get_upper_directories_standard {
  my $self = shift;

  # Current dir
  push @{$self->{upper_dirs}}, {
    name => './ (Current Directory)',
    path => $self->{_scan_fs},
    date => $self->get_date($self->{_scan_fs}),
    type => 'dir',
    size => $self->get_dir_size($self->{_scan_fs}),
    desc => $self->get_dir_desc($self->{_scan_fs}),
    icon => $self->get_dir_icon,
  };

  # Hacky but fast and it works :)
  # Make current directory's icon 'dir-open.png'
  $self->{upper_dirs}->[0]->{icon} =~ s/dir\.png/dir-open\.png/;

  my $entry = dirname $self->{_scan_fs};

  # Parent dir
  push @{$self->{upper_dirs}}, {
    name => '../ (Upper Directory)',
    path => $entry,
    date => $self->get_date($entry),
    type => 'dir',
    size => $self->get_dir_size($entry),
    desc => $self->get_dir_desc($entry),
    icon => $self->get_dir_icon,
    link => $self->get_dir_link($entry),
  };

  1;
}

sub get_entry_info {
  my ($self, $entry) = @_;

  return unless defined $entry;

  # Initialize info hash
  my $info = {};

  # Common stuff
  $info->{path} = $entry;
  $info->{date} = $self->get_date($entry);

  if (-d $entry) {
    # Directory information
    $info->{type} = 'dir';
    $info->{name} = basename($entry) . '/';
    $info->{size} = $self->get_dir_size($entry);
    $info->{desc} = $self->get_dir_desc($entry);
    $info->{icon} = $self->get_dir_icon;
    $info->{link} = $self->get_dir_link($entry);
  } else {
    # File information
    $info->{type} = 'file';
    $info->{name} = basename $entry;
    $info->{size} = $self->get_file_size($entry);
    $info->{desc} = $self->get_file_desc($entry);
    $info->{icon} = $self->get_file_icon($entry);
    $info->{link} = $self->get_file_link($entry);
  }

  return $info;
}

sub print_stuff {
  my $self = shift;

  return unless defined $self->{contents};
  return unless ref $self->{contents} eq 'ARRAY';
  return unless scalar @{$self->{contents}};

  # Use $q for easy access to CGI object
  my $q = $self->{q};

  print $q->start_table({ -class => 'browse', -summary => 'Browse files' });

  print $q->start_thead;

  # Column names to make everything obvious
  my @columns = ('Name', 'Description', 'Size', 'Last Changed');

  # Add another column for icons if we have a mime_images directory
  unshift @columns, '' if defined $self->{_mime_images};

  # This browser's name
  print $q->Tr($q->td({ -colspan => scalar @columns, -class => 'browser-name' }, $self->{table_title})) if ($self->{table_title});

  print $q->Tr($q->td(\@columns));

  print $q->end_thead;

  # Odd/even counter
  my $n = 0;

  # Upper directories, $d for directory
  foreach my $d (@{$self->{upper_dirs}}) {
    my $icon = '';
    $icon = $q->img({ -src => $d->{icon}, alt => 'dir'}) if $d->{icon};

    my $prefix = '';
    $prefix = '&nbsp;' x ($n * 2) if $self->{_mode} eq 'tree';

    my $link = $prefix;
    if ($d->{link}) {
      $link .= $q->a({ -href => $d->{link}}, $d->{name});
    } else {
      $link .= $d->{name};
    }

    # Get row's class ready
    my $row_class = '';
    if ($self->{_mode} eq 'standard') {
      # Standard mode
      if ($d->{name} =~ m'^\./\s+') {
	# Current directory
	$row_class = 'dot-directory';
      } elsif ($d->{name} =~ m'^\.\./\s+') {
	# Parent directory
	$row_class = 'dotdot-directory';
      }
    } elsif ($self->{_mode} eq 'tree') {
      # Tree mode
      if ($n == $#{$self->{upper_dirs}}) {
	# Deepest directory, which is the current dir
	$row_class = 'current-directory';
      } else {
	# upper-directory-odd or upper-directory-even
	$row_class = 'upper-directory-' . ($n % 2 == 0 ? 'even' : 'odd');
      }
    }

    my @row_data = (
      $q->td({ -class => 'name' }, $link),
      $q->td({ -class => 'info' }, $d->{desc}),
      $q->td({ -class => 'size' }, $d->{size}),
      $q->td({ -class => 'date' }, $d->{date}),
    );

    # Prepend icon only if icons are used (checks _mime_images)
    unshift @row_data, $q->td({ -class => 'icon' }, $icon) if $self->{_mime_images};

    print $q->Tr({ -class => $row_class }, @row_data);

    $n++;
  }

  # $e for entry
  foreach my $e (@{$self->{contents}}) {
    my $icon = '';
    $icon = $q->img({ -src => $e->{icon}, alt => $e->{type}}) if $e->{icon};

    my $link = $q->a({ -href => $e->{link}}, $e->{name});

    # row class is {file,dir}-{even,odd}
    my $row_class = $e->{type} . '-' . ($n % 2 == 0 ? 'even' : 'odd');

    my @row_data = (
      $q->td({ -class => 'name' }, $link),
      $q->td({ -class => 'info' }, $e->{desc}),
      $q->td({ -class => 'size' }, $e->{size}),
      $q->td({ -class => 'date' }, $e->{date}),
    );

    # Prepend icon only if icons are used (checks _mime_images)
    unshift @row_data, $q->td({ -class => 'icon' }, $icon) if $self->{_mime_images};

    print $q->Tr({ -class => $row_class }, @row_data);

    $n++;
  }

  # Mode switcher
  print $q->Tr({ -class => 'switcher' }, $q->td({ -colspan => scalar @columns }, $self->mode_switcher)) unless $self->{-no_switcher};

  print $q->end_table;
}

sub _cgi_vars {
  ['d', 'm']
}

sub _valid_module_options {
  return {
    -root_dir     => 1,
    -template     => 1,
    -webmaster    => 1,
    -web_base     => 1,
    -web_base_rel => 1,
    -mime_images  => 1,
    -default_mode => 1,
    -no_switcher  => 1,
  } 
}

sub _required_module_options {
  ['-root_dir', '-template', '-webmaster', '-web_base']
}

sub _valid_display_modes {
  {
    # name => tooltip
    standard => $Texitheque::CGI::StockMessages::ls_Mode_File_Listing_Tip,
    tree     => $Texitheque::CGI::StockMessages::Mini_Tree_Mode_File_Listing_Tip,
  }
}

sub _default_mode {
  'standard'
}

sub _sort_dirs_files {
  if ($a->{type} ne $b->{type}) {
    $a->{type} eq 'dir' ? -1 : 1;
  } else {
    $a->{'name'} cmp $b->{'name'};
  }
}

1;
