package Texitheque::CGI::FileUploader;

use strict;
use warnings;
use base 'Texitheque::CGI';

use CGI::Carp;
use File::Spec;

sub _cgi_vars {
  ['name', 'email', 'desc', 'file', 'Submit']
}

sub _valid_module_options {
  return {
    -contrib_queue    => 1,
    -template         => 1, 
    -webmaster        => 1,
    -archive          => 1,
    -max_post_size    => 1,
    -moderators       => 1,
    -always_notify    => 1,
    -file_count_limit => 1,
    -notify_body      => 1,
    -notify_subject   => 1,
    -desc_file_body   => 1,
    -html_form        => 1,
  }
}

sub _required_module_options {
  ['-template', '-webmaster', '-contrib_queue', '-moderators']
}

sub _valid_display_modes {
}

sub _default_mode {
}

sub contains_swear {
  my ($self, $t) = @_;
  return unless defined $t;
  return $t =~ m/fuck|shit/i;
}

sub process_params {
  my $self = shift;

  my $cgi_err = $self->{q}->cgi_error;

  # Check for '413 POST too large' errors
  if (!$self->{Submit} and defined $cgi_err and $cgi_err =~ m/^413 /) {
    return [ $self->message($Texitheque::CGI::StockMessages::File_Too_Big, { 'maxsize' => $self->deuglify_size($self->{_max_post_size}) })];
  }

  # Do not bother unless the form is submitted
  return unless $self->{Submit};

  my $errors = $self->validate_params;
  return $errors if ref $errors eq 'ARRAY' and scalar @$errors gt 0;

  # We must have a writable $queue directory ready
  unless (-d $self->{contrib_queue}) {
    unless (mkdir $self->{contrib_queue}) {
      warn "Cannot create '$self->{contrib_queue}': $!\n";
      return [$self->message($Texitheque::CGI::StockMessages::Cannot_Save_Uploaded_File, {'error' => $!})];
    }
  }

  my $file   = $self->{q}->upload('file');
  my $output = $self->increment_file(File::Spec->catfile($self->{contrib_queue}, $file));

  # There were already lots of files with the same name
  unless (defined $output) {
    return [$self->message($Texitheque::CGI::StockMessages::Too_Many_Files_In_Queue, {'count' => $self->{_file_count_limit}})];
  }

  if (open(OUT, ">$output")) {
    # Ok, file saved
    print OUT <$file>;
    close(OUT);
    # File should be group writable
    chmod 0664, $output;
  } else {
    warn "Cannot open '$output' for writing: $!\n";
    return [$self->message($Texitheque::CGI::StockMessages::Cannot_Save_Uploaded_File, {'error' => $!})];
  }

  # Get the $file.DSC done
  unless ($self->write_desc($output)) {
    # Unlink the output file if there is trouble
    unlink $output or warn "Cannot unlink($output): $!\n";
    return [$self->message($Texitheque::CGI::StockMessages::Cannot_Save_DSC_File, {'error' => $!})];
  }

  # File part of the new file (may differ with $file due to incremented file)
  my $saved_file = (File::Spec->splitpath($output))[2];

  # Use this file to avoid mailing multiple times
  my $stamp_file = File::Spec->catfile($self->{contrib_queue}, '.notify-stamp');

  # Did we already mail the moderators?
  return 1 if -r $stamp_file and not $self->{always_notify};

  # Notify moderators about the new content
  unless ($self->notify_moderators($file, $saved_file)) {
    return [$self->message($Texitheque::CGI::StockMessages::Cannot_Notify_Moderators, {'webmaster' => $self->{webmaster}})];
  }

  # "We mailed the moderators, so don't bother them again" file
  unlink $stamp_file;
  open my $stamp, '>', $stamp_file or warn "Cannot create time stamp: $!\n";
  chmod 0660, $stamp_file;
  close $stamp;

  return 1;
}

sub validate_params {
  my $self = shift;

  # Save $q->cgi_error as it may get reset during next few checks
  my $file_err = $self->{q}->cgi_error;

  # Check if all the form vars are set
  unless (defined $self->{name} and defined $self->{email} and defined $self->{desc} and defined $self->{file}) {
    return [$Texitheque::CGI::StockMessages::Mangled_Form]
  }

  # Collect errors in this array
  my $errors = [];

  # Validate name
  unless ($self->{name} ne '') {
    push @$errors, $Texitheque::CGI::StockMessages::No_Name_Field;
  } else {
    if ($self->contains_swear($self->{name})) {
      push @$errors, $Texitheque::CGI::StockMessages::Unallowed_Phrase_In_Name;
      $self->{q}->delete('name');
    }
  }

  # Validate email
  unless ($self->{email} ne '') {
    push @$errors, $Texitheque::CGI::StockMessages::No_Email_Field;
  } else {
    unless ($self->check_email) {
      push @$errors, $Texitheque::CGI::StockMessages::Invalid_Email;
      $self->{q}->delete('email');
    }
  }

  # Validate description
  # NOTE: empty values are passed as valid
  if ($self->{desc} ne '') {
    if ($self->contains_swear($self->{desc})) {
      push @$errors, $Texitheque::CGI::StockMessages::Unallowed_Phrase_In_Desc;
      $self->{q}->delete('desc');
    }
  }

  # Validate file
  if ($self->{file} eq '') {
    push @$errors, $Texitheque::CGI::StockMessages::No_File_Uploaded;
  } else {
    if ($file_err) {
      push @$errors, $Texitheque::CGI::StockMessages::Malformed_Post;
    } else {
      my $upload = $self->{q}->upload('file');
      if ((stat($upload))[7] eq 0) {
	push @$errors, $Texitheque::CGI::StockMessages::Empty_Upload;
	$self->{q}->delete('file');
      } elsif ($upload =~ /\.DSC$/) {
	push @$errors, $Texitheque::CGI::StockMessages::Uploaded_DSC_Extension;
	$self->{q}->delete('file');
      } elsif ($upload =~ /^\./) {
	push @$errors, $Texitheque::CGI::StockMessages::Uploaded_Dot_File;
	$self->{q}->delete('file');
      }
    }
  }

  # There were errors, return them
  return $errors if scalar @$errors gt 0;

  # No errors, return nothing
  return;
}

sub check_email {
  my $self = shift;

  return unless defined $self->{email};

  # Checks taken from Mail::CheckUser by Ilya Martynov

  my ($username, $hostname) = $self->{email} =~ /^(.*)@(.*)$/;

  return unless defined $username and defined $hostname;

  # build hostname regexp
  # NOTE: it doesn't strictly follow RFC822
  # because of what registrars now allow.
  my $DOMAIN_RE   = qr/(?:[\da-zA-Z]+ -+)* [\da-zA-Z]+/x;
  my $HOSTNAME_RE = qr/^ (?:$DOMAIN_RE \.)+ [a-zA-Z]+ $/xo;

  return unless $hostname =~ $HOSTNAME_RE;

  # build username regexp
  # NOTE: it doesn't strictly follow RFC821
  my $STRING_RE = ('[' . quotemeta(join '',
				   grep(!/[<>()\[\]\\\.,;:\@"]/, map chr, 33 .. 126)) . ']');
  my $USERNAME_RE = qr/^ (?:$STRING_RE+ \.)* $STRING_RE+ $/xo;

  return unless $username =~ $USERNAME_RE;

  1;
}

sub increment_file {
  my ($self, $file) = @_;

  return unless defined $file;

  # Return itself if it _and_ its DSC file don't exist
  return $file unless -r $file or -r $file . '.DSC';

  my $count = 1;

  while ($count < 10) {
    $count++;
    my $temp_name = $file . '.' . $count;
    last unless -r $temp_name or -r $temp_name . '.DSC';
  }

  return if $count >= $self->{_file_count_limit};

  return $file . '.' . $count;
}

sub html_desc {
  my $self = shift;

  my $desc = $self->{desc};

  return unless defined $desc;

  # Get the carriage returns deleted;
  $desc =~ s/\x0d//g;
  
  # Remove leading and trailing newlines
  $desc =~ s/^\n*//g;
  $desc =~ s/\n*$//g;

  # Substitute < and > with entities
  $desc =~ s/&/\&amp;/g;
  $desc =~ s/</\&lt;/g;
  $desc =~ s/>/\&gt;/g;

  # Remove successive newlines
  # $desc =~ s/\n\n+/\n\n/g;

  # Paragraphize
  $desc = join("\n", map { $_ = '<p>' . $_ . '</p>' } split /\n\n+/, $desc);

  return $desc;
}

sub write_desc {
  my ($self, $output) = @_;

  return unless defined $output;

  my $desc_file = $output . '.DSC';

  unless(open(DESC, ">$desc_file")) {
    warn "Cannot open '$desc_file' for writing: $!\n";
    return;
  }

  my $vars = {
    'name'  => $self->{name},
    'email' => $self->{email},
    'desc'  => $self->{desc},
    'file'  => $output,
    'html_desc' => $self->html_desc,
  };

  print DESC $self->message($self->{_desc_file_body}, $vars);
  close(DESC);
  # File should be group writable
  chmod 0664, $desc_file;
  return 1;
}

sub notify_moderators {
  my ($self, $file, $saved) = @_;

  # We must get all the parameters set
  return unless defined $file and defined $saved and defined $self->{_moderators};

  my $descriptive_file = $saved eq $file ? $saved : "$saved (uploaded as $file)";

  my $vars = {
    'name'  => $self->{name},
    'email' => $self->{email},
    'desc'  => $self->{desc},
    'file'  => $saved,
    'original_file'    => $file,
    'descriptive_file' => $descriptive_file,
  };

  my $subject = $self->message($self->{_notify_subject}, $vars);
  my $body    = $self->message($self->{_notify_body},    $vars);

  unless (open(MAIL, qq{|mail -s '$subject' $self->{_moderators}})) {
    warn "Cannot open pipe to 'mail': $!\n";
    return;
  }

  print MAIL $body;
  close(MAIL);
  return 1;
}

sub queue_size_at_limit {
  my $self = shift;

  my $disk_usage = qx{du -sb '$self->{contrib_queue}' 2>/dev/null};

  unless (defined $disk_usage) {
    warn "Could not run du on '$self->{contrib_queue}': $!\n";
    return;
  }

  chomp $disk_usage;

  $disk_usage =~ /^([0-9]+)/;

  my $size = $1;

  unless (defined $size) {
    warn "Could not get directory size from '$disk_usage'\n";
    return;
  }

  return $size >= $self->{_max_post_size};
}

sub all_ok {
  my $self = shift;

  my $vars = {
    'webmaster' => $self->{webmaster},
    'file' => $self->{q}->upload('file'),
    'archive' => $self->{archive},
  };

  push @{$self->{message_stack}}, $self->message($Texitheque::CGI::StockMessages::All_OK, $vars);

  return 1;
}

sub start_page {
  my $self = shift;

  # Substitute previously sent vars
  my $vars = {};
  my $cgi_vars = $self->_cgi_vars;
  foreach (@$cgi_vars) {
    $vars->{$_} = defined $self->{$_} ? $self->{$_} : '';
  }

  push @{$self->{message_stack}},
    $self->{q}->start_multipart_form,
    $self->message($self->{_html_form}, $vars),
    $self->{q}->end_form;

  return 1;
}

sub got_errors {
  my ($self, $errors) = @_;

  return unless ref $errors eq 'ARRAY';

  push @{$self->{message_stack}},
    join('',
      $self->{q}->div({-class => 'errorbox'},
	$self->{q}->h2({-class => 'error'}, 'Error!'),
	$self->{q}->strong({-class => 'error'}, $Texitheque::CGI::StockMessages::Has_Errors),
	$self->{q}->ul($self->{q}->li($errors)),
      ));

  $self->start_page;
}

sub initialize_vars {
  my $self = shift;

  $self->{message_stack} = [];
  $self->{contrib_queue} = $self->{'-contrib_queue'};
  $self->{always_notify} = $self->{'-always_notify'} ? 1 : 0;

  $self->{_file_count_limit} = defined $self->{'-file_count_limit'} ? $self->{'-file_count_limit'} : 10;

  $self->{archive} = '';
  { 
    my $arc_target = (keys %{$self->{'-archive'}})[0];
    my $arc_title  = (values %{$self->{'-archive'}})[0];
    if (defined $arc_title and defined $arc_target) {
      $self->{archive} = $self->{q}->a({-href=>$arc_target}, $arc_title);
    }
  }

  # Convert from KBytes to Bytes
  $self->{_max_post_size} = $self->{'-max_post_size'} * 1024;

  # Get moderators list ready for mailing
  $self->{_moderators} = join ' ', map { $_ = "'$_'" } @{$self->{'-moderators'}};

  # Get notify email, etc. templates ready
  if (defined $self->{'-notify_subject'}) {
    $self->{_notify_subject} = $self->{'-notify_subject'};
  } else {
    $self->{_notify_subject} = $Texitheque::CGI::StockMessages::Notify_Email_Subject;
  }

  if (defined $self->{'-notify_body'}) {
    $self->{_notify_body} = $self->{'-notify_body'};
  } else {
    $self->{_notify_body} = $Texitheque::CGI::StockMessages::Notify_Email_Body;
  }

  if (defined $self->{'-desc_file_body'}) {
    $self->{_desc_file_body} = $self->{'-desc_file_body'};
  } else {
    $self->{_desc_file_body} = $Texitheque::CGI::StockMessages::Desc_File_Body;
  }

  if (defined $self->{'-html_form'}) {
    $self->{_html_form} = $self->{'-html_form'};
  } else {
    $self->{_html_form} = $Texitheque::CGI::StockMessages::File_Send_Form;
  }
}

sub gather_stuff {
  my $self = shift;

  $self->initialize_vars;

  # Don't do anything if the queue is full
  if ($self->queue_size_at_limit) {
    push @{$self->{message_stack}}, $Texitheque::CGI::StockMessages::Contrib_Queue_Full;
    return 1;
  }

  my $status = $self->process_params;
  $status = -1 unless defined $status;

  if ($status == 1) {
    # Everything is OK
    $self->all_ok;
  } elsif (ref $status eq 'ARRAY') {
    # Something happened with the parameters
    $self->got_errors($status);
  } else {
    # New here?
    $self->start_page;
  }

  return 1;
}

sub print_stuff {
  my $self = shift;

  return unless ref $self->{message_stack} eq 'ARRAY' and scalar @{$self->{message_stack}} > 0;

  print @{$self->{message_stack}};
}

1;
